import os, sys

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))

import cx_Oracle
import pytest

from sa_gwdata import Well, UnitNumber

import sageodata_db

from utils import on_state_intranet

INTRANET = on_state_intranet()


@pytest.mark.skipif(INTRANET is False, reason="Need to be connected to intranet")
def test_test_alive():
    db = sageodata_db.connect()
    assert db.test_alive()


def test_wrong_service_name():
    with pytest.raises(KeyError):
        db = sageodata_db.connect(service_name="DEWNR.World")


@pytest.mark.skipif(INTRANET is False, reason="Need to be connected to intranet")
def test_create_well_instances_dtype():
    db = sageodata_db.connect()
    well = db._create_well_instances([8721])[0]
    assert isinstance(well, Well)


@pytest.mark.skipif(INTRANET is False, reason="Need to be connected to intranet")
def test_create_well_instances_unit_no_parsed():
    db = sageodata_db.connect()
    well = db._create_well_instances([8721])[0]
    assert isinstance(well.unit_no, UnitNumber)


@pytest.mark.skipif(INTRANET is False, reason="Need to be connected to intranet")
def test_create_well_instances_unit_pulled_yep():
    db = sageodata_db.connect()
    well = db._create_well_instances([8721])[0]
    assert well.unit_hyphen == "5930-1049"


@pytest.mark.skipif(INTRANET is False, reason="Need to be connected to intranet")
def test_create_well_instances_easting():
    db = sageodata_db.connect()
    well = db._create_well_instances([8721])[0]
    assert well.easting > 530000 and well.easting < 540000



import socket


def on_state_intranet():
    """Check whether we are on the intranet. This won't work for DEM..."""
    try:
        ip = socket.gethostbyname("darcy")
    except socket.gaierror:
        return False
    else:
        return True



from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

from sageodata_db.utils import *
from sageodata_db.config import *
from sageodata_db.connection import *

import warnings

warnings.filterwarnings("ignore", module="pandas.io.sql", lineno=758)



project_name = "sageodata_db"

project = f"{project_name} documentation"
copyright = "2019-20 DEW"
author = "Kent Inverarity"

from pkg_resources import get_distribution

release = get_distribution(project_name).version
version = ".".join(release.split(".")[:2])

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
]

templates_path = ["_templates"]

source_suffix = ".rst"
master_doc = "index"

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = None

html_theme = "sphinx_rtd_theme"

html_theme_options = {}

html_context = {
    "github_user": "dew-waterscience",
    "github_repo": project_name,
    "github_version": "main",
    "doc_path": "docs",
}

html_static_path = ["_static"]

html_css_files = [
    "css/custom.css",
]


html_sidebars = {
    # "**": ["search-field", "sidebar-nav-bs", "sidebar-ethical-ads"]
    # "**": ["search-field", "sidebar-nav-bs",]
    "**": [
        "search-field",
        "navbar-nav",
        "parent_links",
    ]
}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


def setup(app):
    app.add_css_file("custom.css")



from pathlib import Path
import time
import pyautogui

pyautogui.useImageNotFoundException()

from PIL import Image

# path = Path(__file__).parent

# DEM = Image.open(path / "dem.png")
# DH_NO_LABEL = Image.open(path / "dh_no_label.png")
# NEXT_RECORD = Image.open(path / "next_record.png")
# DELETE_RECORD = Image.open(path / "delete_record.png")

while True:
    try:
        x, y = pyautogui.locateCenterOnScreen("dem.png", grayscale=True)
    except:
        pass
    else:
        # try:
        #     x2, y2 = pyautogui.locateCenterOnScreen("gpsds.png", grayscale=True)
        # except:
        #     pass
        # else:
        pyautogui.click(x=x, y=y)
        x, y = pyautogui.locateCenterOnScreen("delete_record.png", grayscale=True)
        pyautogui.click(x=x, y=y)
        pyautogui.hotkey("ctrl", "s")
        time.sleep(1)

    x, y = pyautogui.locateCenterOnScreen("parent_label.png", grayscale=True)
    pyautogui.click(x=x - 60, y=y)

    try:
        x, y = pyautogui.locateCenterOnScreen("next_record.png", grayscale=True)
    except:
        break
    else:
        pyautogui.click(x=x, y=y)



import os, sys

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))

import pytest

from sageodata_db import utils


def test_chunk():
    assert list(utils.chunk([1, 2, 3], 2)) == [[1, 2], [3]]



from datetime import date, time, datetime, timedelta
import os
from inspect import signature, Parameter
from pathlib import Path
import pprint
from string import Formatter
import sys
import traceback

import cx_Oracle
import pandas as pd
from PIL import Image
from sa_gwdata import *

from .utils import (
    SQL,
    apply_well_id,
    cleanup_columns,
    chunk,
    get_logger,
    parse_query_metadata,
)
from .config import *

from sageodata_db.utils import get_logger


logger = get_logger()

for test_path in (r"C:\Program Files\Oracle\Instant Client v19.9",):
    if Path(test_path).is_dir():
        try:
            cx_Oracle.init_oracle_client(test_path)
        except cx_Oracle.ProgrammingError:
            logger.info(
                "Ignoring cx_Oracle.ProgrammingError - already initialised likely"
            )


def load_predefined_query(filename, stub="queries"):
    if not Path(filename).is_file():
        filename = Path(__file__).parent / stub / filename
    if not Path(filename).is_file():
        return load_predefined_query(str(filename) + ".sql")
    with open(str(filename)) as f:
        query = f.read()
    return query


def get_predefined_query_filenames():
    return [p.name for p in (Path(__file__).parent / "queries").glob("*.sql")]


def get_predefined_query_lookup_types():
    results = {}
    root_path = Path(__file__).parent
    for path in (root_path / "queries").rglob("*"):
        if path.suffix.endswith(".sql"):
            lookup_type = path.parent.name
            if lookup_type == "queries":
                query_content = load_predefined_query(path.name)
                m = re.search(r"\{[a-zA-Z_]*\}", query_content)
                if m:
                    lookup_type = m.group()
                else:
                    lookup_type = ""
            if not lookup_type in results:
                results[lookup_type] = [path.name]
            else:
                results[lookup_type].append(path.name)
    return results


def connect(user="gwquery", password="gwquery", **kwargs):
    """Connect to SA Geodata.

    Args:
        user (str): oracle user
        password (str): password
        service_name (str): version of SA Geodata you would like to connect
            to - options are "DMED.WORLD" or "dev"; "DMET.WORLD" or "test" or
            "QA"; or "DMEP.WORLD" or "prod" - see
            :func:`sageodata_db.config.normalize_service_name` for details.

    Other keyword arguments are passed to
    :func:`sageodata_db.config.make_connection_string`.

    Returns:
        a :class:`sageodata_db.SAGeodataConnection` object.

    Example:

        >>> from sageodata_db import connect
        >>> db = connect()
        >>> db
        <sageodata_db.connection.SAGeodataConnection to gwquery@pirsapd07.pirsa.sa.gov.au:1521/DMEP.World>

    """
    connect_string = make_connection_string(**kwargs)
    return SAGeodataConnection(user=user, password=password, dsn=connect_string)


class SAGeodataConnection(cx_Oracle.Connection):
    """:class:`cx_Oracle.Connection` with methods to read groundwater data.

    This should not be instantiated directly. Instead use::

        >>> from sageodata_db import connect
        >>> db = connect()
        >>> type(db)
        <class 'sageodata_db.connection.SAGeodataConnection'>

    """

    SQL = SQL

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        from sageodata_db import __version__

        self.__version__ = __version__
        self._all_replacement_drillholes = self.all_replacement_drillholes()

    def _make_predefined_query_lambda(self, query_name):
        return lambda *args, **kwargs: self._predefined_query(
            query_name, *args, **kwargs
        )

    def __repr__(self):
        return super().__repr__()

    def test_alive(self):
        """Test whether connection is functioning.

        Returns:
            bool: ``True`` or ``False``

        """
        df = self.query("select distinct owner from all_tables")
        try:
            owners = df["owner"]
        except cx_Oracle.DatabaseError:
            return False
        if "SYS" in owners.values:
            return True
        return False

    def _predefined_query(
        self,
        filename,
        *args,
        stub="queries",
        query_method="query",
        include_replacements=False,
        **kwargs,
    ):
        """Run a predefined query on the database.

        Args:
            query_name (str): see below for possible predefined queries.

        Other args and kwargs are passed to `_predefined_query()` along with the fixed
        kwargs `stub="queries", query_method="query"`.

        Returns:
            pandas DataFrame.

        """
        sql = load_predefined_query(filename, stub=stub)
        query_method = getattr(self, query_method)
        df = query_method(SQL(sql, *args, **kwargs))
        cols = [x for x in df.columns]
        if (
            "obs_no" in df.columns
            and "unit_hyphen" in df.columns
            and "dh_no" in df.columns
        ):
            if len(df):
                df["well_id"] = df.apply(apply_well_id, axis="columns")
            else:
                df = df.assign(well_id=None)
            cols = ["well_id"] + cols
        df = df[cols]
        logger.debug(f"Include replacements? {include_replacements}")
        if include_replacements:
            dfs = []
            for dh_no in df.dh_no.unique():
                all_dhs = self.find_replacements([dh_no])
                if len(all_dhs) > 0:
                    missing_dhs = set(all_dhs.old_dh_no).difference(set(df.dh_no))
                    logger.debug(f"missing_dhs {missing_dhs}")
                    all_df = self._predefined_query(
                        filename,
                        list(missing_dhs),
                        include_replacements=False,
                        **kwargs,
                    )
                    df2 = pd.concat([df, all_df])
                    if "obs_date" in df2.columns:
                        dt_col = "obs_date"
                    elif "collected_date" in df2.columns:
                        dt_col = "collected_date"

                    df3 = df2[df2.dh_no == all_dhs.iloc[0].old_dh_no]

                    for idx, row in all_dhs.sort_values("replaced_from").iterrows():
                        logger.debug(
                            f"Adding in data from replacement: {row.to_dict()}"
                        )
                        if ~pd.isnull(row.new_dh_no) and ~pd.isnull(row.replaced_from):
                            # Remove measurements from the old_dh_no after or on the replacement date.
                            remove_old = (df3.dh_no == row.old_dh_no) & (
                                df3[dt_col] >= row.replaced_from
                            )
                            df3 = df3[~remove_old]
                            # Add measurements from the new_dh_no on and after the replacement date.
                            keep_new = (df2.dh_no == row.new_dh_no) & (
                                df2[dt_col] >= row.replaced_from
                            )
                            df3 = pd.concat([df3, df2[keep_new]])
                    df3 = df3.sort_values([dt_col])

                    # Improve columns
                    current = (
                        self.drillhole_details(all_dhs.index.unique()).iloc[0].to_dict()
                    )
                    logger.debug(f"Current well is: {current}")
                    well_cols = [
                        "well_id",
                        "dh_no",
                        "unit_long",
                        "unit_hyphen",
                        "obs_no",
                        "dh_name",
                        "easting",
                        "northing",
                        "zone",
                        "latitude",
                        "longitude",
                        "aquifer",
                    ]
                    for col in ["dh_no", "unit_long", "unit_hyphen", "obs_no"]:
                        df3[f"from_{col}"] = df3[col]
                    for col in well_cols[::-1]:
                        df3[col] = current[col]
                else:
                    df3 = df[df.dh_no == dh_no]
                dfs.append(df3)
            df = pd.concat(dfs)
        return df

    def query(self, query, lowercase_columns=True, *args, conn=None, **kwargs):
        """Run an SQL query on SA Geodata and return a :class:`pandas.DataFrame`.

        Args:
            query (str, SQL): SQL query as string or
                :class:`sageodata_db.connection.SQL`

        Returns:
            a :class:`pandas.DataFrame` object.

        """
        if conn is None:
            conn = self
        if isinstance(query, str):
            query = SQL(query, *args, **kwargs)
        dfs = []
        for sql in query:
            logger.debug("running query:\n{}".format(sql))
            dfs.append(pd.read_sql_query(sql, conn))
        df = pd.concat(dfs)
        if lowercase_columns:
            df = df.rename(columns=str.lower)
        return df

    def _create_well_instances(self, dh_nos):
        """Private method to create Well instances from dh_nos.

        Args:
            dh_nos (array-like): drillhole numbers as integers.

        Returns:
            :class:`sa_gwdata.Wells` object (sequence of :class:`sa_gwdata.Well` objects).

        """
        df = self.drillhole_details(dh_nos)
        df.loc[:, "name"] = df["dh_name"]
        df.loc[:, "unit_no.hyphen"] = df["unit_hyphen"]
        df.loc[:, "unit_no"] = df["unit_hyphen"]
        df.loc[:, "dh_no"] = df["dh_no"]
        return Wells([Well(**vals.to_dict()) for _, vals in df.iterrows()])

    def find_replacement_history(self, dh_no):
        """Find the chain of drillholes that replaced one another, for *dh_no*.

        Args:
            dh_no (int): drillhole number

        Returns:
            a pandas dataframe with columns "old_dh_no", "new_dh_no", and
            "replaced_from".

        """

        df = self._all_replacement_drillholes

        def _find_replacement_history(
            dh_no_chain, check_for_newer_wells=True, check_for_older_wells=True
        ):
            earliest_dh_no = dh_no_chain[0]
            latest_dh_no = dh_no_chain[-1]
            if check_for_newer_wells:
                replacements = df[df["dh_no"] == latest_dh_no]
                if len(replacements) == 1:
                    row = replacements.iloc[0]
                    dh_no_chain = list(dh_no_chain) + [row.new_dh_no]
                    return _find_replacement_history(
                        dh_no_chain, check_for_newer_wells=True
                    )
            if check_for_older_wells:
                replaced = df[df["new_dh_no"] == earliest_dh_no]
                if len(replaced) == 0:
                    return dh_no_chain
                elif len(replaced) == 1:
                    row = replaced.iloc[0]
                    previous_dh_no = row.dh_no
                    dh_no_chain = [previous_dh_no] + list(dh_no_chain)
                    return _find_replacement_history(
                        dh_no_chain, check_for_newer_wells=False
                    )

        dh_no_chain = _find_replacement_history([dh_no])
        hdf = df[df.dh_no.isin(dh_no_chain)]
        hdf = hdf.set_index("dh_no")
        hdf2 = hdf.reindex(dh_no_chain)
        hdf3 = hdf2.reset_index().rename(columns={"dh_no": "old_dh_no"})
        return hdf3

    def find_replacements(self, wells):
        """Find replacements for a set of wells.

        Args:
            wells (Wells, pandas dataframe, or list of ints): either must
                have an attribute "dh_no", or a list of drillhole numbers.

        Returns:
            pandas DataFrame with index "current_dh_no" and
            columns "queried_dh_no", "old_dh_no", "new_dh_no", "replaced_from"

        """
        if hasattr(wells, "dh_no"):
            dh_nos = [getattr(o, "dh_no") for o in wells]
        else:
            dh_nos = wells
        dfs = []
        for dh_no in dh_nos:
            df = self.find_replacement_history(dh_no)
            if len(df) > 1:
                df.insert(0, "current_dh_no", _find_current_for_replacement_chain(df))
                df.insert(0, "queried_dh_no", dh_no)
            else:
                df = df.assign(current_dh_no=dh_no, queried_dh_no=dh_no)
            dfs.append(df)
        return pd.concat(dfs).set_index("current_dh_no")

    def find_wells(self, input_text, **kwargs):
        """Find wells and retrieve some summary information.

        Args:
            input_text (str): any well identifiers to parse. See
                :func:`sa_gwdata.parse_well_ids_plaintext` for details of
                other keyword arguments you can pass here.

        Returns:
            wells (pandas.DataFrame): a table of well summary
                details (via :meth:`sageodata_db.SAGeodataConnection.wells_summary`)

        For example::

            >>> from sageodata_db import connect
            >>> db = connect()
            >>> wells = db.find_wells(": G662801265, 6628-14328, YAT30, and ULE 205.")

        See :func:`sa_gwdata.parse_well_ids` for more details.

        """
        ids = parse_well_ids(input_text, **kwargs)
        dh_nos = [x for id_type, x in ids if id_type == "dh_no"]
        unit_nos = [x for id_type, x in ids if id_type == "unit_no"]
        obs_nos = [x for id_type, x in ids if id_type == "obs_no"]
        unit_long_ints = [UnitNumber(x).long_int for x in unit_nos]
        logger.debug("unit_no.long_int: {}".format(unit_long_ints))
        r_obs_nos = []

        if obs_nos:
            r1 = self._predefined_query("drillhole_no_by_obs_no.sql", obs_nos)
            r_obs_nos = r1.dh_no.tolist()
        r_unit_nos = []
        if unit_long_ints:
            r2 = self._predefined_query("drillhole_no_by_unit_long.sql", unit_long_ints)
            r_unit_nos = r2.dh_no.tolist()
        logger.debug("obs_nos -> {}".format(r_obs_nos))
        logger.debug("unit_nos -> {}".format(r_unit_nos))
        all_dh_nos = list(set(dh_nos + r_obs_nos + r_unit_nos))
        return self.wells_summary(all_dh_nos)
        # return self._create_well_instances(all_dh_nos)

    def find_wells_from_df(
        self,
        df,
        copy=True,
        return_id_cols=("unit_hyphen",),
        remap_id_col_names=True,
        **kwargs,
    ):
        """Find wells identifier in a dataframe and merge them together.

        Args:
            df (pd.DataFrame): dataframe containing well IDs
            copy (bool): if True return a copy otherwise just return well IDs.
            return_id_cols (list of str): columns containing identified well
                IDs which are returned. If *remap_id_col_names* is True, you can
                pick from: dh_no, well_id, dh_name, unit_hyphen, unit_long, unit_wilma
                unit_hydstra, obs_no, and obs_egis. If *remap_id_col_names* is False,
                you can choose from dh_no, id, name, unit_no.map, unit_no.seq,
                unit_no.hyphen, unit_no.long_int, unit_no.long,
                unit_no.wilma, unit_no.hydstra, obs_no.plan, obs_no.seq, obs_no.id,
                and obs_no.egis.
            remap_id_col_names (bool): remap from Wells.df to the typical columns
                found in other data queries.

        Other kwargs are passed to parse_well_ids().

        Returns:
            If **copy** == ``True``, a dataframe with **return_id_cols** added at the start, with all
            well identifiers matched. Otherwise, a dataframe with only **return_id_cols** is
            provided.

        """
        if isinstance(df, pd.Series):
            df = df.to_frame()

        # Parse any well identifier for each row of the
        # dataframe.
        wids = df.apply(
            lambda series: parse_well_ids("|".join(map(str, series.values)), **kwargs),
            axis="columns",
        )

        # Use only the first of each well identifier in each row
        wids = [wid[0] if len(wid) else "" for wid in wids]

        search_lines = list(set([str(wid[1]) if len(wid) == 2 else "" for wid in wids]))
        vwids_dfs = []
        for search in chunk(search_lines, 1000):
            logger.info(f"Searching for {len(search)} well IDs")
            vwids_dfs.append(self.find_wells("/".join(search), **kwargs))
        vwids = pd.concat(vwids_dfs).drop_duplicates()
        values = []
        for wid in wids:
            row = False
            if wid == "":
                pass
            elif wid[0] == "unit_no":
                row_sel = vwids[vwids["unit_hyphen"] == wid[1]]
                if len(row_sel) == 1:
                    row = row_sel.iloc[0]
            elif wid[0] == "obs_no":
                row_sel = vwids[vwids["obs_no"] == wid[1]]
                if len(row_sel) == 1:
                    row = row_sel.iloc[0]
            elif wid[0] == "dh_no":
                row_sel = vwids[vwids.dh_no == wid[1]]
                if len(row_sel) == 1:
                    row = row_sel.iloc[0]
            if row is False:
                row = pd.Series([None for x in vwids.columns], index=vwids.columns)
            row.loc["parsed_id"] = wid
            values.append(row)
        id_cols = pd.DataFrame(values, index=df.index.values)
        if remap_id_col_names:
            id_cols = id_cols.rename(
                columns={
                    "id": "well_id",
                    "name": "dh_name",
                    "unit_no.hyphen": "unit_hyphen",
                    "unit_no.long_int": "unit_long",
                    "unit_no.wilma": "unit_wilma",
                    "unit_no.hydstra": "unit_hydstra",
                    "obs_no.id": "obs_no",
                    "obs_no.egis": "obs_egis",
                }
            )
        if return_id_cols is None:
            return_id_cols = id_cols.columns
        else:
            return_id_cols = ["parsed_id"] + list(return_id_cols)
        if copy:
            return pd.concat([id_cols[return_id_cols], df], axis=1)
        else:
            return id_cols[return_id_cols].T.drop_duplicates().T

    def find_edits_by(self, *modified_by):
        """Return a table of edits made to certain tables by a set of users.

        The tables which are current queried are water levels, salinity samples,
        and elevation surveys.

        Args:
            *modified_by (str): SA Geodata usernames

        Returns:
            A pandas dataframe.

        """
        wls = self.water_level_edits(modified_by)
        sals = self.salinity_edits(modified_by)
        elevs = self.elevation_edits(modified_by)
        wls = _remove_elev_mods_from_wl(wls, elevs)
        wls = wls.groupby(["well_id", "modified_date"]).filter(lambda x: len(x) <= 1)
        df = pd.concat(
            [
                wls.assign(table="WL").rename(columns={"wl": "data_values"}),
                sals.assign(table="Salinity").rename(
                    columns={"salinity": "data_values"}
                ),
                elevs.assign(table="Elevation").rename(columns={"elev": "data_values"}),
            ]
        ).sort_values("modified_date", ascending=False)
        cols = [x for x in df.columns if not x == "table"]
        cols.insert(cols.index("well_id") + 1, "table")
        return df[cols]

    def find_additions_by(self, *created_by):
        """Return a table of additions made to certain tables by a set of users.

        The tables which are current queried are water levels, salinity samples,
        and elevation surveys.

        Args:
            *created_by (str): SA Geodata usernames

        Returns:
            A pandas dataframe.

        """
        wls = self.water_level_additions(created_by)
        sals = self.salinity_additions(created_by)
        elevs = self.elevation_additions(created_by)
        wls = wls.groupby(["well_id", "creation_date"]).filter(lambda x: len(x) <= 1)
        df = pd.concat(
            [
                wls.assign(table="WL").rename(columns={"wl": "data_values"}),
                sals.assign(table="Salinity").rename(
                    columns={"salinity": "data_values"}
                ),
                elevs.assign(table="Elevation").rename(columns={"elev": "data_values"}),
            ]
        ).sort_values("creation_date", ascending=False)
        cols = [x for x in df.columns if not x == "table"]
        cols.insert(cols.index("well_id") + 1, "table")
        return df[cols]

    def open_drillhole_document_image(self, image_no):
        """Load a drillhole document image from DD_DH_IMAGE

        Args:
            image_no (int): see predefined query "drillhole_document_image_list"

        Returns:
            ``PIL.Image`` object

        """
        cursor = self.cursor()
        var = cursor.var(cx_Oracle.BLOB)
        cursor.execute(
            """
                declare
                    t_Image ordsys.ordimage;
                begin
                    select Image
                    into t_Image
                    from dhdb.dd_dh_image where image_no = {:.0f};

                    :1 := t_Image.source.localdata;

                end;""".format(
                image_no
            ),
            (var,),
        )
        blob = var.getvalue()
        return Image.open(blob)
    
    def open_drillhole_image(self, image_no):
        """Load a drillhole image (photograph) from IM_IMAGE

        Args:
            image_no (int): see predefined query "drillhole_image_list"

        Returns:
            ``PIL.Image`` object

        """
        cursor = self.cursor()
        var = cursor.var(cx_Oracle.BLOB)
        cursor.execute(
            """
                declare
                    t_Image ordsys.ordimage;
                begin
                    select Image
                    into t_Image
                    from dhdb.im_image where image_no = {:.0f};

                    :1 := t_Image.source.localdata;

                end;""".format(
                image_no
            ),
            (var,),
        )
        blob = var.getvalue()
        return Image.open(blob)

    def open_db_file(self, file_no):
        """Open database file.

        Args:
            file_no (int)

        Returns:
            A tuple, where the first item is a string with the filename, the second is a ``io.BytesIO`` object
            with the file contents.

        """
        cursor = self.cursor()
        var = cursor.var(cx_Oracle.BLOB)
        cursor.execute(
            f"select file_name, file_contents from dhdb.fi_file where file_no = {file_no}"
        )
        for item in cursor:
            file_name = item[0]
            file_object = item[1]
        return file_name, io.BytesIO(file_object.read())

    def open_db_file_as_text(
        self, file_no, encoding=("ascii", "cp1252", "utf-8"), raise_error=True
    ):
        """Open database file as text.

        Args:
            file_no (int)
            encoding (sequence of str or str): the encodings to try in the order to try them.

        Returns:
            A tuple. The first item is a string with the filename, the second is a string
            with the file contents.

        """
        filename, buffer = self.open_db_file(file_no)
        buffer_contents = buffer.read()
        if isinstance(encoding, str):
            encoding = [encoding]

        contents = False
        i = 0
        while i < len(encoding):
            try_encoding = encoding[i]
            try:
                contents = buffer_contents.decode(try_encoding)
                contents = contents.strip()
                if len(contents) > 0:
                    break
            except:
                i += 1
        if contents is False:
            if raise_error:
                try:
                    raise
                except:
                    contents = ""
            else:
                logger.warning(f"Encoding error reading file_no={file_no}")
                contents = ""
        return filename, contents

    def list_geophys_log_db_files(self, job_nos=None):
        """Query for a list of geophysical log files stored in SA Geodata.

        Args:
            job_nos (sequence of int, optional): filter to particular
                job numbers, otherwise this will return all records in SAGD.

        The query is hard-coded into this function and is across FI_FILE,
        FI_FILE_LINK_VW and GL_LOG_HDR_VW.

        Returns:
            pandas.DataFrame: The table has columns:
            
            - job_no (int)
            - path (str). This column does not contain file paths, but is
              used for compatibility with functions utilising the gtslogs
              shared folder location. In this case it is populated with
              e.g. "sagd:file_no=1234"
            - file_type (str): filename suffix in uppercase
            - file_size (float): not populated as the files are not
              retrieved.

        """
        if job_nos:
            df = self.geophys_log_files_by_job_no(job_nos)
        else:
            df = self.geophys_log_files_all()

        df = df.rename(columns={"file_name": "filename", "file_no": "path"})
        df = df.drop(["log_hdr_no"], axis=1)
        df["path"] = df.path.apply(lambda file_no: f"sagd:file_no={file_no}")
        df["file_type"] = df.filename.apply(lambda fn: Path(fn).suffix.upper()[1:])
        return df

    def lookup_unit_numbers(conn, unit_nos):
        """Look up unit numbers.

        Args:
            unit_nos (sequence of str or int): anything that can be parsed by
                sa_gwdata.UnitNumber

        Returns:
            pandas.DataFrame: This has columns dh_no, unit_long,
            unit_hyphen, obs_no, dh_name etc.

        Order is not preserved, sorry.
        
        """
        std_unit_nos = [UnitNumber(u) for u in unit_nos]
        mapsheet_sets = {}
        for u in std_unit_nos:
            if not u.map in mapsheet_sets:
                mapsheet_sets[u.map] = [u.seq]
            else:
                mapsheet_sets[u.map].append(u.seq)
        dfs = []
        for mapsheet_num, dh_seq_nos in mapsheet_sets.items():
            dfs.append(conn.drillhole_by_dh_seq_no(mapsheet_num, dh_seq_nos))
        df = pd.concat(dfs)
        return df
    
    def lookup_obs_numbers(conn, obs_nos):
        """Look up obswell IDs.

        Args:
            obs_nos (sequence of str): anything that can be parsed by
                sa_gwdata.ObsNumber

        Returns:
            pandas.DataFrame: This has columns dh_no, unit_long,
            unit_hyphen, obs_no, dh_name etc.

        Order is not preserved, sorry.
        
        """
        std_obs_nos = [ObsNumber(o) for o in obs_nos]
        hundred_sets = {}
        for o in std_obs_nos:
            if not o.plan in hundred_sets:
                hundred_sets[o.plan] = [o.seq]
            else:
                hundred_sets[o.plan].append(o.seq)
        dfs = []
        for hundred_sets, dh_seq_nos in hundred_sets.items():
            dfs.append(conn.drillhole_by_obs_seq_no(hundred_sets, dh_seq_nos))
        df = pd.concat(dfs)
        return df
    
    def aquifer_units_details(self, aquifer_code):
        df = self.aquifer_units(aquifer_code)
        p = re.compile(r"\[SU(\d*)\]")
        
        def extract_su(desc):
            m = p.search(desc)
            if m:
                return int(m.group(1))
            else:
                return pd.NA
                
        out_cols = [
                "aquifer_code", "hydro_subunit_desc", "linked_strat_unit_no", "linked_map_symbol", "linked_strat_name", "linked_strat_agso_number", "major_strat_unit_no", "major_map_symbol", "hydro_subunit_code" 
                ]
        if len(df) == 0:
            return pd.DataFrame(columns=out_cols)
        else:
            df = df.rename(columns={
                "strat_unit_no": "major_strat_unit_no",
                "map_symbol": "major_map_symbol",
                "strat_name": "major_strat_name"
            })
            df["linked_strat_unit_no"] = df.hydro_subunit_desc.apply(extract_su)
            su = self.strat_unit_details(df.linked_strat_unit_no.dropna())
            su = su.rename(columns={
                "strat_unit_no": "linked_strat_unit_no",
                "map_symbol": "linked_map_symbol",
                "strat_name": "linked_strat_name",
                "agso_number": "linked_strat_agso_number",
            })
            su = su[["linked_strat_unit_no", "linked_map_symbol", "linked_strat_name", "linked_strat_agso_number"]]
            dfm = pd.merge(df, su, on="linked_strat_unit_no")
            dfm = dfm[out_cols]
            return dfm

    


def _remove_elev_mods_from_wl(wl, elev):
    """Remove any edit records from WL data when there is an edit made to
    the elevation table within 5 seconds.

    """
    remove_indices = []
    wl_copy = pd.DataFrame(wl)
    for (dh_no, modified_date), gdf in elev.groupby(["dh_no", "modified_date"]):
        n1 = len(wl[wl.dh_no == dh_no])
        date1 = modified_date - timedelta(seconds=2)
        date2 = modified_date + timedelta(seconds=2)
        remove_idx = (
            (wl.dh_no == dh_no)
            & (wl.modified_date >= date1)
            & (wl.modified_date <= date2)
        )
        n2 = len(wl[remove_idx])
        logger.debug(f"{dh_no}, {modified_date}: wl {n1} -> {n2}")
        remove_indices.append(remove_idx)
        wl_copy = wl_copy[~remove_idx]
    return wl_copy


def _find_current_for_replacement_chain(df):
    """Identify which drillhole is current in a chain of
    replacement drillholes.

    Args:
        df (pandas DataFrame): returned by
            SAGeodataConnection.find_replacement_history()

    Returns:
        int, the drillhole number of the current well.

    """
    candidate = df[df.new_dh_no.isnull()].iloc[0]
    candidate_start = df.loc[df.new_dh_no == candidate.old_dh_no, "replaced_from"].iloc[
        0
    ]
    other_candidates = df.loc[df.replaced_from > candidate_start, :]
    if len(other_candidates) > 0:
        return int(other_candidates.sort_values("replaced_from").iloc[-1].new_dh_no)
    else:
        return int(candidate.old_dh_no)


def __monkey_patch_predefined_queries(cls, method_name):
    """Monkey-patch predefined query methods on to a class.

    Args:
        cls (class): class to monkey-patch methods on to
        method_name (str): name of method in cls which takes
            the query_name as its first argument. Other args
            and kwargs are passed through to this method

    Returns:
        nothing.

    """

    def __make_lambda(query_name, method_name):
        return lambda self, *args, **kwargs: getattr(self, method_name)(
            query_name, *args, **kwargs
        )

    docstring_lines = []
    query_names = get_predefined_query_filenames()
    for query_name in query_names:
        query = load_predefined_query(query_name)
        columns, param_defns, docstring_parts = parse_query_metadata(query)
        query_name = query_name.replace(".sql", "")
        func = __make_lambda(query_name, method_name=method_name)
        func.__name__ = query_name
        field_names = [fname for _, fname, _, _ in Formatter().parse(query) if fname]
        if field_names:
            params = [Parameter("self", Parameter.POSITIONAL_OR_KEYWORD)]
            for field_name in field_names:
                param_meta = param_defns.get(field_name, {"default": None, "descr": ""})
                if not param_meta["default"]:
                    param_meta["default"] = Parameter.empty
                params.append(
                    Parameter(
                        field_name,
                        Parameter.POSITIONAL_OR_KEYWORD,
                        default=param_meta["default"],
                    )
                )
            if "dh_no" in [n.lower() for n in field_names]:
                params.append(
                    Parameter(
                        "include_replacements",
                        Parameter.POSITIONAL_OR_KEYWORD,
                        default=False,
                    )
                )
                param_defns["include_replacements"] = {
                    "descr": (
                        "include_replacements (bool): default ``False``. "
                        "Fold data from replaced drillholes the newest "
                        "drillhole (e.g. data from dh_no 12345 will be included"
                        " as part of drillhole 23500 if 23500 is marked as a "
                        "replacement for 12345)"
                    ),
                    "default": False,
                }

            unique_params = []
            for p in params:
                if not p in unique_params:
                    unique_params.append(p)
            func.__signature__ = signature(func).replace(parameters=unique_params)

        doclines = [f"Run predefined query '{query_name}.sql'.", ""]
        if docstring_parts[0]:
            doclines += docstring_parts[0]
            doclines += [""]
        if param_defns:
            doclines += ["Args:"]
            doclines += [
                f"    {meta['descr'].strip()}" for name, meta in param_defns.items()
            ]
            doclines += [""]
        if len(columns):
            doclines += ["Returns:", "    :class:`pandas.DataFrame`", ""]
            doclines += ["The returned DataFrame has the following columns:", ""]
            doclines += [
                f" - ``'{name}'`` {meta['descr']}" for name, meta in columns.items()
            ]
            doclines += [""]
        if docstring_parts[1]:
            doclines += docstring_parts[1]
            doclines += [""]
        # doclines += [
        #     "The SQL template (used by :class:`sageodata_db.SQL`) for this query:",
        #     "",
        #     ".. code-block:: plpgsql",
        #     "",
        # ]
        # doclines += [
        #     f"    {line}"
        #     for line in query.splitlines()
        #     if not line.strip().startswith("-- ")
        # ]
        # doclines += [""]

        func.__doc__ = "\n".join([f"    {line}" for line in doclines])
        setattr(cls, query_name, func)


__monkey_patch_predefined_queries(SAGeodataConnection, method_name="_predefined_query")



import os, sys

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))

import cx_Oracle
import pytest

from sa_gwdata import Well, UnitNumber

import sageodata_db
from sageodata_db import get_predefined_query_filenames


from utils import on_state_intranet

INTRANET = on_state_intranet()


def get_predefined_query_filenames():
    assert "hydrostrat_logs.sql" in get_predefined_query_filenames()


@pytest.mark.skipif(INTRANET is False, reason="Need to be connected to intranet")
def test_drillhole_details():
    db = sageodata_db.connect()
    assert 593001049 in db.drillhole_details([8721]).unit_long.values


@pytest.mark.skipif(INTRANET is False, reason="Need to be connected to intranet")
def test_hydrostrat_logs_result_length():
    db = sageodata_db.connect()
    assert len(db.hydrostrat_logs([8721, 201270])) == 9



from setuptools import setup

setup(
    name="sageodata_db",
    packages=["sageodata_db"],
    package_data={"sageodata_db": ["queries/*.sql"]},
    use_scm_version={"version_scheme": "post-release"},
    setup_requires=["setuptools_scm"],
    description="SA Geodata database querying tool/wrapper/library",
    long_description=open("README.md", mode="r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/dew-waterscience/sageodata_db",
    author="DEW Water Science (Kent Inverarity)",
    author_email="groundwater@sa.gov.au",
    license="All rights reserved",
    classifiers=("Programming Language :: Python :: 3",),
    keywords="science",
    install_requires=(
        "cx_Oracle",
        "pandas",
        "python-sa-gwdata",
        "loguru",
        "sqlparse",
        "pillow",
    ),
)



from datetime import datetime, date
from string import Formatter
from pathlib import Path
import re
import sys
import warnings

warnings.filterwarnings("ignore")

import numpy as np
import pandas as pd
import sqlparse
from sa_gwdata import Well, Wells


def get_logger():
    from loguru import logger

    # logger.remove()
    # logger.add(sys.stdout, level="WARNING")
    return logger


logger = get_logger()


class SQL:
    r"""Represents an SQL query or queries.

    Args:
        sql (str): SQL query template
        to_str (bool): TBD
        chunksize (int): when sequences are passed into template fields,
            this is the maximum length allowed before a new query is issued.
        ignore_fields (sequence of str): list of keyword arguments to ignore.

    Remaining positional and keyword arguments are assumed to be filling
    template fields in ``sql``, unless they are listed in ``ignore_fields``.

    Lower-case string template fields are filled directly with the content of the
    keyword argument, while upper-case ones are understood as iterators over a variety
    of sequence types (each then represented in the Oracle SQL correctly according to the
    sequence's element data type). Data-types are processed into SQL according
    to the following tests:

     - if a sequence contains Python ints, they are represented as integers
       in the generated SQL
     - if a sequence contains Python floats, they are represented as floating
       point numbers in the generated SQL
     - if a sequence contains Python datetime, date, or :class:`pandas.Timestamp`
       objects, they are represented as strings in the generated SQL with the
       Python datetime formatting ``%Y-%m-%d %H:%M:%S``.
     - all other Python objects are represented as strings in the generated
       SQL

    The generated SQL should always be retrieved by iterating over this object,
    even if sequences are not being passed. An example of how this class can be
    used follows.

        >>> from sageodata_db import SQL
        >>> query = SQL(
        ...     "select * from dhdb.dd_drillhole_vw where drillhole_no in {dh_no}",
        ...     dh_no=1
        ... )
        >>> for sql in query:
        ...     print(sql)
        SELECT *
        FROM dhdb.dd_drillhole_vw
        WHERE drillhole_no IN 1

    Or for a sequence data type:

        >>> from sageodata_db import SQL
        >>> sequence_query = SQL(
        ...     "select * from dhdb.dd_drillhole_vw where drillhole_no in {DH_NO}",
        ...     dh_no=[1, 2, 3, 4, 5, 6, 7]
        ... )
        >>> for sql in sequence_query:
        ...     print(sql)
        SELECT *
        FROM dhdb.dd_drillhole_vw
        WHERE drillhole_no IN (1,2,3,4,5,6,7)

    To illustrate how a long list of qualifiers is automatically broken into the
    maximum acceptable length by the database engine, let's artifically reduce
    the default chunk size of 1000 to something we can easily visualize:

        >>> sequence_query.chunksize = 3
        >>> for i, sql in enumerate(sequence_query):
        ...     print((i, sql))
        (0, 'SELECT *\nFROM dhdb.dd_drillhole_vw\nWHERE drillhole_no IN (1,2,3)')
        (1, 'SELECT *\nFROM dhdb.dd_drillhole_vw\nWHERE drillhole_no IN (4,5,6)')
        (2, 'SELECT *\nFROM dhdb.dd_drillhole_vw\nWHERE drillhole_no IN (7)')

    The kwarg to_str provides a function which turns the elements from field_list
    to a string. By default it is determined by the type of the first element.

    You can re-use a :class:`sageodata_db.SQL` object with a new field_list:

        >>> sequence_query_2 = SQL(sequence_query, [8, 9, 10])
        >>> for sql in sequence_query_2:
        ...     print(sql)
        SELECT *
        FROM dhdb.dd_drillhole_vw
        WHERE drillhole_no IN (8,9,10)

    """

    def __init__(
        self, sql, *args, to_str=None, chunksize=1000, ignore_fields=None, **kwargs
    ):
        if isinstance(sql, SQL):
            sql = sql.sql

        if ignore_fields is None:
            ignore_fields = ()

        self.chunksize = chunksize
        self.sql = sqlparse.format(sql, reindent=True, keyword_case="upper")

        fields = [
            fname
            for _, fname, _, _ in Formatter().parse(self.sql)
            if fname and (not fname in ignore_fields) and (not re.match(r"\d+", fname))
        ]

        # Assign a field name to each positional argument. Most of the time there
        # will only be one positional argument, and only one field in the SQL query.
        # But in general, we assign them in the order we find them, and clobber
        # anything from **kwargs in the process.

        for i in range(len(args)):
            kwargs[fields[i]] = args[i]

        # Fields in uppercase are to be filled as lists.
        # Fields in lowercase are to be filled as items for every query.

        uppercase_fields = [x for x in fields if x.upper() == x]
        lowercase_fields = [x for x in fields if not x in uppercase_fields]
        # assert len(uppercase_fields) in (0, 1)

        # logger.debug(f"uppercase_fields: {uppercase_fields}")
        # logger.debug(f"lowercase_fields: {lowercase_fields}")

        # If an SQL field e.g. DH_NO is present in kwargs in lowercase, then we need
        # to convert the kwargs to uppercase, so that everything else works sensibly.

        for upper_field in uppercase_fields:
            keys = list(kwargs.keys())
            for k in keys:
                if k == upper_field.lower():
                    kwargs[upper_field] = kwargs[k]
                    del kwargs[k]
                    break

        if len(uppercase_fields) > 0:
            items = kwargs[uppercase_fields[0]]
            if isinstance(items, Wells):
                items = getattr(
                    items, uppercase_fields[0].lower()
                )  # e.g. for {DH_NO}, fetch [w.dh_no for w in Wells]
            elif isinstance(items, pd.DataFrame):
                items = items[uppercase_fields[0].lower()].tolist()
            self.field_list = items
            self.field_list_name = uppercase_fields[0]  # remain uppercase
        else:
            self.field_list = []
            self.field_list_name = None

        self.to_str_funcs = {}
        for field_name, example in kwargs.items():
            if field_name == field_name.upper():
                if isinstance(example, pd.DataFrame):
                    example = example.iloc[0]
                else:
                    try:
                        example = example[0]
                    except IndexError:
                        example = None

            # np.int64 needs conversion to int
            if "numpy.int" in str(type(example)):
                example = int(example)

            if example is None:
                # Field list is empty. We need a valid SQL query, so that we
                # return an empty table with the correct column names.
                # We assume that nothing will match an empty string in the
                # SQL where clause.

                self.to_str_funcs[field_name] = lambda x: "'{}'".format(str(x))
                self.field_list = [""]

            else:
                if isinstance(example, int) or isinstance(example, np.int64):
                    self.to_str_funcs[field_name] = lambda x: str(int(x))
                elif isinstance(example, float):
                    self.to_str_funcs[field_name] = lambda x: str(float(x))
                elif (
                    isinstance(example, datetime)
                    or isinstance(example, pd.Timestamp)
                    or isinstance(example, date)
                ):
                    self.to_str_funcs[field_name] = lambda x: x.strftime(
                        "'%Y-%m-%d %H:%M:%S'"
                    )
                else:
                    if "fragment" in field_name.lower():
                        target = "'%{}%'"
                    else:
                        target = "'{}'"

                    if "uppercase" in field_name.lower():
                        self.to_str_funcs[field_name] = lambda x: target.format(
                            str(x).upper()
                        )
                    else:
                        self.to_str_funcs[field_name] = lambda x: target.format(str(x))

        if self.field_list_name:
            del kwargs[self.field_list_name]

        self.scalar_fields = kwargs

    def __iter__(self):
        scalar_inserts = {
            k: self.to_str_funcs[k](v) for k, v in self.scalar_fields.items()
        }

        if len(self.field_list):
            for sub_list in chunk(self.field_list, self.chunksize):
                to_str = self.to_str_funcs[self.field_list_name]
                sub_list_str = "(" + ",".join(map(to_str, sub_list)) + ")"
                inserts = dict(scalar_inserts)
                inserts[self.field_list_name] = sub_list_str
                result = self.sql.format(**inserts)
                yield result
        elif len(scalar_inserts):
            result = self.sql.format(**scalar_inserts)
            yield result
        else:
            result = self.sql
            yield result


PREDEF_COL_SPECS = {
    "dh_no": "drillhole number (unique ID for drillholes) - integer",
    "unit_long": "- unit number in 9-digit numeric integer form e.g. 662800123",
    "unit_hyphen": "- unit number in condensed hyphenated form e.g. 6628-123",
    "obs_no": "- observation well number in 6-character padded form e.g. YAT056",
    "dh_name": "- drillhole name",
    "easting": "- UTM easting in metres, GDA2020",
    "northing": "- UTM northing in metres, GDA2020",
    "zone": "- UTM zone, GDA2020",
    "latitude": "- geographic latitude in GDA2020 datum as decimal degrees",
    "longitude": "- geographic longitude in GDA2020 datum as decimal degrees",
    "aquifer": "- the current aquifer code - e.g. the combination of strat symbol and hydro subunit such as 'Tomw(T1)'. If there are multiple aquifer monitored codes, they are joined with a plus e.g. 'Qpah+Tomw(T1)'",
    "created_by": "- SA Geodata user that orginally created the record. DHDB is a system user.",
    "creation_date": "- Date and time that the record was originally created.",
    "modified_by": "- SA Geodata user that most recently modified the record.",
    "modified_date": "- date and time that the record was most recently modified (:class:`pandas.Timestamp`)",
}


def parse_query_metadata(query):
    columns = {}
    params = {}
    docstring = [[], []]
    first_part_over = False
    for line in query.splitlines():
        if line.startswith("--") and not first_part_over:
            docstring[0].append(line.strip("-- "))
        elif line.startswith("--") and first_part_over:
            docstring[1].append(line.strip("-- "))
        else:
            first_part_over = True
            if "--col" in line:
                before, after = line.split("--col", 1)
                name = before.split()[-1]
                name = name.strip(",")
                if "." in name:
                    name = name.split(".")[-1]
                descr = after.strip()
                if descr:
                    descr = " - " + descr
                elif name in PREDEF_COL_SPECS.keys():
                    descr = PREDEF_COL_SPECS[name]
                columns[name] = {"descr": descr}
            elif "--arg" in line:
                descr = line.split("--arg")[-1]
                parts = descr.split()
                if len(parts) > 1:
                    name = parts[0]
                else:
                    name = "?"
                m = re.search(r"default=(['\".a-zA-Z0-9]+)", descr)
                if m:
                    default = m.group(1)
                else:
                    default = None
                params[name] = {"descr": descr, "default": default}
    return columns, params, docstring


def chunk(l, n=1000):
    """Yield successive n-sized chunks from a list l.

    >>> from sageodata_db.utils import chunk
    >>> for x in chunk([0, 1, 2, 3, 4], n=2):
    ...     print(x)
    [0, 1]
    [2, 3]
    [4]

    """
    y = 0
    for i in range(0, len(l), n):
        y += 1
        yield l[i : i + n]
    if y == 0:
        yield l


def apply_well_id(row, columns=("obs_no", "unit_hyphen", "dh_no")):
    for col in columns:
        if row[col]:
            return row[col]
    return ""


def cleanup_columns(df, keep_cols="well_id", remove_metadata=False):
    """Remove unneeded drillhole identifier columns.

    Args:
        df (pandas DataFrame): dataframe to remove columns from
        keep_cols (sequence of str): columns to retain (only applies to the
            well identifiers columns; any other columns will be retained
            regardless)

    Returns: dataframe

    """
    if not "well_id" in df.columns:
        cols = [x for x in df.columns]
        df["well_id"] = df.apply(apply_well_id, axis="columns")
        df = df[["well_id"] + cols]
    if remove_metadata:
        for col in df:
            if (
                "modified_date" in col
                or "creation_date" in col
                or "modified_by" in col
                or "created_by" in col
            ):
                df = df.drop(col, axis=1)
    keep_columns = []
    for col in df.columns:
        if (
            col
            in (
                "well_id",
                "dh_no",
                "unit_long",
                "unit_hyphen",
                "obs_no",
                "dh_name",
                "easting",
                "northing",
                "zone",
                "latitude",
                "longitude",
                "aquifer",
            )
            and not col in keep_cols
        ):
            pass
        else:
            keep_columns.append(col)
    return df[keep_columns]



import cx_Oracle

from sageodata_db.utils import get_logger


logger = get_logger()

PIRSA_DEV_SERVER = "pirzapd08.pirsa.sa.gov.au"
PIRSA_TEST_SERVER = "pirzapd08.pirsa.sa.gov.au"
PIRSA_PROD_SERVER = "pirsapd07.pirsa.sa.gov.au"
DEW_DEV_SERVER = "sageodata.db.dev.env.sa.gov.au"
DEW_TEST_SERVER = "sageodata.db.dev.env.sa.gov.au"
DEW_PROD_SERVER = ""

PORT = 1521


def normalize_service_name(name: str) -> str:
    """Ensure consistent SA Geodata service name.

    Args:
        service_name (str): database service name either
            "prod", "test" or "dev" or
            one of "DMET.WORLD@PIRSA", "DMED.WORLD@PIRSA",
            "DMEP.WORLD@PIRSA", "DMET@DEW", "DMED@DEW" or
            "DMEP.WORLD@DEW". If the @PIRSA or @DEW is
            omitted @PIRSA will be used.

    Returns:
        string: The proper service name e.g. DMEP.WORLD@PIRSA
        for the current "prod".

    """
    name = name.upper()

    if not "@" in name:
        name += "@PIRSA"

    name1, name2 = name.split("@")
    name1 = name1.strip().upper()
    name2 = name2.strip().upper()

    if name1 == "PROD":
        name1 = "DMEP.WORLD"
    elif name1 == "TEST" or name1.lower() == "QA":
        name1 = "DMET.WORLD"
    elif name1 == "DEV":
        name1 = "DMED.WORLD"

    error = "service@location - "
    if not name1.upper() in ("DMEP.WORLD", "DMET.WORLD", "DMED.WORLD"):
        raise KeyError(error + "service must be either PROD, TEST, or DEV")
    if not name2 in ("DEW", "PIRSA"):
        raise KeyError(error + "location must be either DEW or PIRSA")

    if name2 == "DEW":
        if "DMET" in name1:
            name1 = "DMET"
        elif "DMED" in name1:
            name1 = "DMED"

    return name1 + "@" + name2


def find_appropriate_server(service_name: str) -> str:
    """Find the server that each service name lives on. See PROD_SERVER
    and TEST_SERVER.

    Args:
        service_name (str): one of "DMET.WORLD@PIRSA", "DMED.WORLD@PIRSA",
            "DMEP.WORLD@PIRSA", "DMET@DEW", "DMED@DEW" or "DMEP.WORLD@DEW".

    Returns: 
        string: Server address.

    """
    if service_name == "DMET.WORLD@PIRSA":
        server = PIRSA_TEST_SERVER
    elif service_name == "DMED.WORLD@PIRSA":
        server = PIRSA_DEV_SERVER
    elif service_name == "DMEP.WORLD@PIRSA":
        server = PIRSA_PROD_SERVER
    elif service_name == "DMET@DEW":
        server = DEW_TEST_SERVER
    elif service_name == "DMED@DEW":
        server = DEW_DEV_SERVER
    elif service_name == "DMEP.WORLD@DEW":
        server = DEW_PROD_SERVER
    else:
        raise KeyError(
            "service_name must be either DMEP.WORLD@DEW, DMET.WORLD@DEW, or DMED.WORLD@DEW (or @PIRSA)"
        )
    return server


def makedsn(service_name="DMEP.WORLD", server=None, port=PORT):
    """Get the appropriate Oracle DSN.

    Args:
        service_name (str): database service name one of
            "DMET.WORLD@PIRSA", "DMED.WORLD@PIRSA", "DMEP.WORLD@PIRSA",
            "DMET@DEW", "DMED@DEW" or "DMEP.WORLD@DEW".
            This goes to :func:`sageodata_db.normalize_service_name`
            first so you can also pass "prod", "test", or "dev".
        server (str, optional): server address
        port (int, optional): port

    Returns:
        string: DSN connection string.

    For example, to get the production database:

        >>> from sageodata_db import makedsn
        >>> makedsn("prod")
        '(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST=pirsapd07.pirsa.sa.gov.au)(PORT=1521))(CONNECT_DATA=(SERVICE_NAME=DMEP.World)))'

    """
    service_name = normalize_service_name(service_name)
    if server is None:
        server = find_appropriate_server(service_name)
    return cx_Oracle.makedsn(server, port, service_name=service_name.split("@")[0])


def make_connection_string(service_name="DMEP.WORLD", server=None, port=PORT):
    """Get the appropriate cx_Oracle connection string.

    Args:
        service_name (str): database service name one of
            "DMET.WORLD@PIRSA", "DMED.WORLD@PIRSA", "DMEP.WORLD@PIRSA",
            "DMET@DEW", "DMED@DEW" or "DMEP.WORLD@DEW".
            This goes to :func:`sageodata_db.normalize_service_name`
            first so you can also pass "prod", "test", or "dev".
        server (str, optional): server address
        port (int, optional): port

    Returns:
        string: DSN connection string.

    For example, to get the production database:

        >>> from sageodata_db import makedsn
        >>> makedsn("prod")
        '(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST=pirsapd07.pirsa.sa.gov.au)(PORT=1521))(CONNECT_DATA=(SERVICE_NAME=DMEP.World)))'

    """
    service_name = normalize_service_name(service_name)
    if server is None:
        server = find_appropriate_server(service_name)
    return "{}:{}/{}".format(server, port, service_name.split("@")[0])
